import os
import shutil
import sys
import unittest
import click
from typing import List

from spaceone.core import config, pygrpc, fastapi, utils, model
from spaceone.core import plugin as plugin_srv
from spaceone.core import scheduler as scheduler_v1
from spaceone.core.unittest.runner import RichTestRunner
from spaceone.core.logger import set_logger
from spaceone.core.opentelemetry import set_tracer, set_metric
from spaceone.core.plugin.plugin_conf import PLUGIN_SOURCES

_GLOBAL_CONFIG_PATH = "{package}.conf.global_conf:global_conf"


@click.group()
def cli():
    pass


@cli.command()
@click.argument("project_name")
@click.option("-d", "--directory", type=click.Path(), help="Project directory")
@click.option(
    "-s",
    "--source",
    type=str,
    help=f"skeleton code of the plugin: ["
    f'{"|".join(PLUGIN_SOURCES.keys())}] or '
    f"module path(e.g. spaceone.core.skeleton)]",
)
def create_project(project_name, directory=None, source=None):
    """Create a new project"""

    _create_project(project_name, directory, source)


@cli.group()
def run():
    """Run a server"""
    pass


@run.command()
@click.argument("package")
@click.option(
    "-a",
    "--app-path",
    type=str,
    help="Python path of gRPC application [default: {package}.interface.grpc:app]",
)
@click.option(
    "-s",
    "--source-root",
    type=click.Path(exists=True),
    default=".",
    help="Path of source root",
    show_default=True,
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=os.environ.get("SPACEONE_PORT", 50051),
    help="Port of gRPC server",
    show_default=True,
)
@click.option(
    "-w",
    "--worker",
    type=int,
    default=os.environ.get("SPACEONE_WORKER", 100),
    help="Worker of gRPC server",
    show_default=True,
)
@click.option(
    "-c",
    "--config-file",
    type=click.Path(exists=True),
    default=os.environ.get("SPACEONE_CONFIG_FILE"),
    help="Path of config file",
)
@click.option(
    "-m",
    "--module-path",
    type=click.Path(exists=True),
    multiple=True,
    help="Additional python path",
)
def grpc_server(
    package,
    app_path=None,
    source_root=None,
    port=None,
    worker=None,
    config_file=None,
    module_path=None,
):
    """Run a gRPC server"""

    # Initialize config
    _set_server_config(
        package,
        source_root,
        port,
        config_file=config_file,
        grpc_app_path=app_path,
        module_path=module_path,
        worker=worker,
    )

    # Initialize common modules
    _init_common_modules()

    # Run gRPC server
    pygrpc.serve()


@run.command()
@click.argument("package")
@click.option(
    "-a",
    "--app-path",
    type=str,
    help="Python path of REST application [default: {package}.interface.rest:app]",
)
@click.option(
    "-s",
    "--source-root",
    type=click.Path(exists=True),
    default=".",
    help="Path of source root",
    show_default=True,
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=os.environ.get("SPACEONE_PORT", 8000),
    help="Port of REST server",
    show_default=True,
)
@click.option(
    "-h",
    "--host",
    type=str,
    default=os.environ.get("SPACEONE_HOST", "127.0.0.1"),
    help="Host of REST server",
    show_default=True,
)
@click.option(
    "-c",
    "--config-file",
    type=click.Path(exists=True),
    default=os.environ.get("SPACEONE_CONFIG_FILE"),
    help="Path of config file",
)
@click.option(
    "-m",
    "--module-path",
    type=click.Path(exists=True),
    multiple=True,
    help="Additional python path",
)
def rest_server(
    package,
    app_path=None,
    source_root=None,
    port=None,
    host=None,
    config_file=None,
    module_path=None,
):
    """Run a FastAPI REST server"""

    # Initialize config
    _set_server_config(
        package,
        source_root,
        port,
        host=host,
        config_file=config_file,
        rest_app_path=app_path,
        module_path=module_path,
    )

    # Initialize common modules
    _init_common_modules()

    # Run REST server
    fastapi.serve()


@run.command()
@click.argument("package")
@click.option(
    "-s",
    "--source-root",
    type=click.Path(exists=True),
    default=".",
    help="Path of source root",
    show_default=True,
)
@click.option(
    "-c",
    "--config-file",
    type=click.Path(exists=True),
    default=os.environ.get("SPACEONE_CONFIG_FILE"),
    help="Path of config file",
)
@click.option(
    "-m",
    "--module-path",
    type=click.Path(exists=True),
    multiple=True,
    help="Additional python path",
)
def scheduler(package, source_root=None, config_file=None, module_path=None):
    """Run a scheduler server"""

    # Initialize config
    _set_server_config(
        package, source_root, config_file=config_file, module_path=module_path
    )

    # Initialize common modules
    _init_common_modules()

    # Run scheduler server
    scheduler_v1.serve()


@run.command()
@click.argument("package")
@click.option(
    "-a",
    "--app-path",
    type=str,
    help="Path of Plugin application [default: {package}.main:app]",
)
@click.option(
    "-s",
    "--source-root",
    type=click.Path(exists=True),
    default=".",
    help="Path of source root",
    show_default=True,
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=os.environ.get("SPACEONE_PORT", 50051),
    help="Port of plugin server",
    show_default=True,
)
@click.option(
    "-w",
    "--worker",
    type=int,
    default=os.environ.get("SPACEONE_WORKER", 100),
    help="Worker of gRPC server",
    show_default=True,
)
@click.option(
    "-m",
    "--module-path",
    type=click.Path(exists=True),
    multiple=True,
    help="Additional python path",
)
def plugin_server(
    package, app_path=None, source_root=None, port=None, worker=None, module_path=None
):
    """Run a plugin server"""

    # Initialize config
    _set_server_config(
        package,
        source_root,
        port,
        plugin_app_path=app_path,
        module_path=module_path,
        set_custom_config=False,
        worker=worker,
    )

    # Run Plugin Server
    plugin_srv.serve()


@cli.command()
@click.argument("package")
@click.option(
    "-c",
    "--config-file",
    type=click.Path(exists=True),
    default=lambda: os.environ.get("SPACEONE_CONFIG_FILE"),
    help="Path of config file",
)
@click.option(
    "-s",
    "--source-root",
    type=click.Path(exists=True),
    default=".",
    help="Path of source root",
    show_default=True,
)
@click.option(
    "-o",
    "--output",
    default="yaml",
    help="Output format",
    type=click.Choice(["json", "yaml"]),
    show_default=True,
)
def show_config(package, source_root=None, config_file=None, output=None):
    """Show global configurations"""
    # Initialize config
    _set_server_config(package, source_root, config_file=config_file)

    # Print merged config
    _print_config(output)


@cli.command()
@click.option("-c", "--config-file", type=str, help="Path of config file")
@click.option(
    "-d",
    "--dir",
    type=str,
    help="Directory containing test files",
    default=lambda: os.environ.get("SPACEONE_WORKING_DIR", os.getcwd()),
)
@click.option("-f", "--failfast", help="Fast failure flag", is_flag=True)
@click.option("-s", "--scenario", type=str, help="Path of scenario file")
@click.option(
    "-p",
    "--parameters",
    type=str,
    help="Custom parameters to override a scenario file. "
    "(e.g. -p domain.domain.name=new_name -p options.update_mode=false)",
    multiple=True,
)
@click.option("-v", "--verbose", count=True, help="Verbosity level", default=1)
def test(
    config_file=None,
    dir=None,
    failfast=False,
    scenario: str = None,
    parameters: List[str] = None,
    verbose=1,
):
    """Unit tests for source code"""
    # set config
    if config:
        os.environ["TEST_CONFIG"] = config_file

    if scenario:
        os.environ["TEST_SCENARIO"] = scenario

    if parameters:
        os.environ["TEST_SCENARIO_PARAMS"] = ",".join(parameters)

    # run test
    loader = unittest.TestLoader()
    suites = loader.discover(dir)

    full_suite = unittest.TestSuite()
    full_suite.addTests(suites)
    RichTestRunner(verbosity=verbose, failfast=failfast).run(full_suite)


def _set_python_path(
    package: str, source_root: str = None, module_path: List[str] = None
):
    source_root = source_root or os.getcwd()

    if source_root not in sys.path:
        sys.path.insert(0, source_root)

    if module_path:
        for path in module_path:
            if path not in sys.path:
                sys.path.insert(0, path)

    try:
        __import__(package)
    except Exception:
        raise Exception(
            f"The package({package}) can not imported. " "Please check the module path."
        )


def _set_server_config(
    package,
    source_root=None,
    port=None,
    host=None,
    config_file=None,
    grpc_app_path=None,
    rest_app_path=None,
    plugin_app_path=None,
    module_path=None,
    set_custom_config=True,
    worker=None,
):
    # 1. Set a python path
    _set_python_path(package, source_root, module_path)

    # 2. Initialize config from command argument
    config.init_conf(
        package=package,
        port=port,
        worker=worker,
        host=host,
        grpc_app_path=grpc_app_path,
        rest_app_path=rest_app_path,
        plugin_app_path=plugin_app_path,
    )

    if set_custom_config:
        # 3. Get service config from global_conf.py
        config.set_service_config()

        # 4. Merge file conf
        if config_file:
            config.set_file_conf(config_file)


def _create_project(project_name, directory=None, source=None):
    # Initialize path
    project_name = project_name
    project_directory = directory or os.getcwd()
    project_path = os.path.join(project_directory, project_name)

    # Copy skeleton source code
    if source:
        skeleton = PLUGIN_SOURCES.get(source, source)
    else:
        skeleton = "spaceone.core.skeleton"

    # Check skeleton module name
    module_name = skeleton.split(".")[-1]
    if module_name != "skeleton":
        raise Exception('Skeleton module path must be ended with "skeleton".')

    # Copy skeleton source code
    skeleton_module = __import__(skeleton, fromlist=["*"])
    skeleton_path = os.path.dirname(skeleton_module.__file__)
    shutil.copytree(
        skeleton_path, project_path, ignore=shutil.ignore_patterns("__pycache__")
    )


def _print_config(output):
    data = {"GLOBAL": config.get_global()}

    if output == "json":
        print(utils.dump_json(data, indent=4))
    else:
        print(utils.dump_yaml(data))


def _init_common_modules() -> None:
    # Enable logging configuration
    set_logger()

    # Set OTel Tracer and Metric
    set_tracer()

    # Connect all databases
    model.init_all()


if __name__ == "__main__":
    cli()



import logging
import traceback

from functools import wraps
from fastapi import Request, HTTPException
from spaceone.core.error import *
from spaceone.core.locator import Locator

_LOGGER = logging.getLogger(__name__)

_STATUS_CODE_MAP = {
    'OK': 200,
    'CANCELLED': 499,
    'UNKNOWN': 500,
    'INVALID_ARGUMENT': 400,
    'DEADLINE_EXCEEDED': 504,
    'NOT_FOUND': 404,
    'ALREADY_EXISTS': 409,
    'PERMISSION_DENIED': 403,
    'UNAUTHENTICATED': 401,
    'FAILED_PRECONDITION': 400,
    'ABORTED': 409,
    'OUT_OF_RANGE': 400,
    'UNIMPLEMENTED': 501,
    'INTERNAL': 500,
    'UNAVAILABLE': 503,
    'DATA_LOSS': 500
}


def exception_handler(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except ERROR_BASE as e:
            _print_error_log(e)
            _raise_exception(e)
        except Exception as e:
            error = ERROR_UNKNOWN(message=str(e))
            _print_error_log(error)
            _raise_exception(error)

    return wrapper


def _print_error_log(error):
    if not error.meta.get('skip_error_log'):
        _LOGGER.error(f'(Error) => {error.message} {error}', exc_info=True)


def _raise_exception(error):
    raise HTTPException(status_code=_check_http_status_code(error.status_code),
                        detail={'code': error.error_code, 'message': error.message})


def _check_http_status_code(grpc_status_code: str) -> int:
    return _STATUS_CODE_MAP.get(grpc_status_code, 500)


class BaseAPI(object):
    locator = Locator()
    service = None

    @property
    def name(self):
        return self.__class__.__name__

    async def parse_request(self, request: Request, token=None, resource=None, verb=None):
        try:
            metadata = {
                'url': request.url.path,
                'service': self.service,
                'resource': resource,
                'verb': verb,
                'token': token
            }

            return await request.json(), metadata
        except Exception as e:
            _LOGGER.debug(f'JSON Parsing Error: {e}', exc_info=True)
            raise ERROR_UNKNOWN(message='JSON Parsing Error: Request body requires JSON format.')



import logging
from spaceone.core.handler import BaseMutationHandler

_LOGGER = logging.getLogger(__name__)


class SpaceONEMutationHandler(BaseMutationHandler):
    def request(self, params):
        user_role_type: str = self.transaction.get_meta("authorization.role_type")
        domain_id: str = self.transaction.get_meta("authorization.domain_id")
        workspace_id: str = self.transaction.get_meta("authorization.workspace_id")
        user_projects: list = self.transaction.get_meta("authorization.projects")
        user_id: str = self.transaction.get_meta("authorization.user_id")
        set_user_id: str = self.transaction.get_meta("authorization.set_user_id")
        injected_params: dict = self.transaction.get_meta(
            "authorization.injected_params"
        )

        if user_role_type == "SYSTEM_TOKEN":
            if domain_id:
                params["domain_id"] = domain_id

            if workspace_id:
                params["workspace_id"] = workspace_id

        elif user_role_type == "DOMAIN_ADMIN":
            params["domain_id"] = domain_id
        elif user_role_type == "WORKSPACE_OWNER":
            params["domain_id"] = domain_id
            params["workspace_id"] = workspace_id
        elif user_role_type == "WORKSPACE_MEMBER":
            params["domain_id"] = domain_id
            params["workspace_id"] = workspace_id
            params["user_projects"] = user_projects or []
        elif user_role_type == "USER":
            params["domain_id"] = domain_id

        if set_user_id:
            params["user_id"] = user_id

        if injected_params:
            params.update(injected_params)

        return params



import abc
import logging
from typing import List
from spaceone.core.base import CoreObject
from spaceone.core import config
from spaceone.core.error import *

__all__ = [
    "BaseHandler",
    "BaseAuthenticationHandler",
    "BaseAuthorizationHandler",
    "BaseMutationHandler",
    "BaseEventHandler",
    "get_authentication_handlers",
    "get_authorization_handlers",
    "get_mutation_handlers",
    "get_event_handlers",
]

_HANDLER_TYPE = ["authentication", "authorization", "mutation", "event"]
_HANDLER_INFO = {
    "init": False,
    "authentication": [],
    "authorization": [],
    "mutation": [],
    "event": [],
}

_LOGGER = logging.getLogger(__name__)


class BaseHandler(CoreObject):
    def __init__(self, handler_config: dict):
        super().__init__()

        self.config = handler_config


class BaseAuthenticationHandler(abc.ABC, BaseHandler):
    @abc.abstractmethod
    def verify(self, params: dict) -> None:
        """
        Args:
            params (dict): Request Parameter

        Returns:
            None
        """
        raise NotImplementedError("verify method not implemented!")


class BaseAuthorizationHandler(abc.ABC, BaseHandler):
    @abc.abstractmethod
    def verify(
        self, params: dict, permission: str = None, role_types: list = None
    ) -> None:
        """
        Args:
            params (dict): Request Parameter
            permission (str): API Permission
            role_types (list): Allowed Role Types

        Returns:
            None
        """
        raise NotImplementedError("verify method not implemented!")


class BaseMutationHandler(BaseHandler):
    def request(self, params: dict) -> dict:
        """
        Args:
            params (dict): Request Parameter

        Returns:
            params (dict): Changed Parameter
        """
        return params

    def response(self, result: any) -> any:
        """
        Args:
            result (dict): Result data

        Returns:
            result (dict): Changed Result Data
        """
        return result


class BaseEventHandler(abc.ABC, BaseHandler):
    @abc.abstractmethod
    def notify(self, status: str, message: dict) -> None:
        """
        Args:
            status (str): 'Transaction Status: STARTED or IN_PROGRESS or SUCCESS or FAILURE'
            message (dict): 'Request Parameter or Response Data or Error Message'

        Returns:
            None
        """
        raise NotImplementedError("notify method not implemented!")


def _init_handlers() -> None:
    registered_handlers = config.get_global("HANDLERS", {})
    for handler_type in _HANDLER_TYPE:
        for handler_conf in registered_handlers.get(handler_type, []):
            try:
                module_name, class_name = handler_conf["backend"].rsplit(":", 1)
                _LOGGER.debug(f"[_init_handlers] {handler_type} handler: {class_name}")

                handler_module = __import__(module_name, fromlist=[class_name])

                _HANDLER_INFO[handler_type].append(
                    handler_module.__dict__[class_name](handler_conf)
                )

            except Exception as e:
                raise ERROR_HANDLER_CONFIGURATION(handler=handler_type, reason=str(e))


def get_authentication_handlers() -> List[BaseAuthenticationHandler]:
    _check_init_state()
    return _HANDLER_INFO.get("authentication", [])


def get_authorization_handlers() -> List[BaseAuthorizationHandler]:
    _check_init_state()
    return _HANDLER_INFO.get("authorization", [])


def get_mutation_handlers(reverse: bool = False) -> List[BaseMutationHandler]:
    _check_init_state()
    if reverse:
        return _HANDLER_INFO.get("mutation", [])[::-1]
    else:
        return _HANDLER_INFO.get("mutation", [])


def get_event_handlers() -> List[BaseEventHandler]:
    _check_init_state()
    return _HANDLER_INFO.get("event", [])


def _check_init_state() -> None:
    if not _HANDLER_INFO["init"]:
        _init_handlers()
        _HANDLER_INFO["init"] = True



import logging
from spaceone.core.transaction import get_transaction


class TransactionFilter(logging.Filter):
    def filter(self, record):
        if transaction := get_transaction(is_create=False):
            record.service = transaction.service
            record.trace_id = transaction.id
            record.domain_id = transaction.get_meta("authorization.domain_id") or ""
            record.audience = transaction.get_meta("authorization.audience") or ""
            record.role_type = transaction.get_meta("authorization.role_type") or ""
            record.workspace_id = (
                transaction.get_meta("authorization.workspace_id") or ""
            )
            record.peer = transaction.get_meta("peer")
            if transaction.resource and transaction.verb:
                record.tnx_method = f"{transaction.resource}.{transaction.verb}"
            else:
                record.tnx_method = ""
        else:
            record.service = ""
            record.trace_id = ""
            record.domain_id = ""
            record.audience = ""
            record.role_type = ""
            record.workspace_id = ""
            record.tnx_status = ""
            record.tnx_method = ""
            record.peer = ""
        return True



import abc


class Authenticator(object):

    def __init__(self, key):
        self._key = key

    @abc.abstractmethod
    def validate(self, token):
        raise NotImplementedError('Valid must be implemented')



class BaseCache(object):

    def get(self, key, **kwargs):
        """
        Args:
            key (str)
            **kwargs (dict)
        Returns:
            cache_value (any)
        """
        raise NotImplementedError('cache.get not implemented!')

    def set(self, key, value, **kwargs):
        """
        Args:
            key (str)
            value (str)
            **kwargs (dict)
                - expire (int: seconds)

        Returns:
            True | False
        """
        raise NotImplementedError('cache.set not implemented!')

    def increment(self, key, amount):
        """
        Args:
            key (str)
            amount (int)

        Returns:
            True | False
        """
        raise NotImplementedError('cache.incr not implemented!')

    def decrement(self, key, amount):
        """
        Args:
            key (str)
            amount (int)

        Returns:
            True | False
        """
        raise NotImplementedError('cache.decr not implemented!')

    def keys(self, pattern='*'):
        """
        Args:
            pattern (str)

        Returns:
            keys (list)
        """
        raise NotImplementedError('cache.keys not implemented!')

    def ttl(self, key):
        """
        Args:
            key (str)

        Returns:
            expire_time (int: seconds)
        """
        raise NotImplementedError('cache.ttl not implemented!')

    def delete(self, *keys):
        """
        Args:
            keys (list)

        Returns:
            None
        """
        raise NotImplementedError('cache.delete not implemented!')

    def delete_pattern(self, pattern):
        """
        Args:
            pattern (str)

        Returns:
            None
        """
        raise NotImplementedError('cache.delete_pattern not implemented!')

    def flush(self, is_async=False):
        """
        Args:
            is_async (bool)

        Returns:
            None
        """
        raise NotImplementedError('cache.flush not implemented!')



import logging

from spaceone.core import config
from spaceone.core.error import *

_LOGGER = logging.getLogger(__name__)


def _get_module(package, target):
    return __import__(f"{package}.{target}", fromlist=[f"{target}"])


class Locator(object):
    @staticmethod
    def get_service(name_or_object: [str, object], metadata: dict = None, **kwargs):
        metadata = metadata or {}
        package = config.get_package()

        if isinstance(name_or_object, str):
            service_module = _get_module(package, "service")
            return getattr(service_module, name_or_object)(metadata=metadata, **kwargs)
        else:
            return name_or_object(metadata=metadata, **kwargs)

    @staticmethod
    def get_manager(name_or_object: [str, object], **kwargs):
        package = config.get_package()
        try:
            if isinstance(name_or_object, str):
                manager_module = _get_module(package, "manager")
                return getattr(manager_module, name_or_object)(**kwargs)
            else:
                return name_or_object(**kwargs)

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(
                name=name_or_object, reason=e, _meta={"type": "manager"}
            )

    @staticmethod
    def get_info(name_or_object: [str, object], *args, **kwargs):
        package = config.get_package()
        try:
            if isinstance(name_or_object, str):
                info_module = _get_module(package, "info")
                return getattr(info_module, name_or_object)(*args, **kwargs)
            else:
                return name_or_object(*args, **kwargs)

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(name=name_or_object, reason=e, _meta={"type": "info"})

    @staticmethod
    def get_model(name_or_object: [str, object]):
        package = config.get_package()
        try:
            if isinstance(name_or_object, str):
                model_module = _get_module(package, "model")
                return getattr(model_module, name_or_object)
            else:
                return name_or_object

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(
                name=f"{name_or_object} Model", reason=e, _meta={"type": "model"}
            )

    @staticmethod
    def get_connector(name_or_object: [str, object], **kwargs):
        package = config.get_package()

        try:
            if isinstance(name_or_object, str):
                connector_conf = config.get_connector(name_or_object)
                backend = connector_conf.get("backend")

                if backend:
                    connector_module_path, connector_name = backend.rsplit(":", 1)
                    connector_module = __import__(
                        connector_module_path, fromlist=[connector_name]
                    )
                else:
                    connector_module = _get_module(package, "connector")
                    connector_name = name_or_object

                return getattr(connector_module, connector_name)(
                    connector_name=name_or_object, **kwargs
                )
            else:
                return name_or_object(**kwargs)

        except ERROR_BASE as e:
            raise e

        except Exception as e:
            raise ERROR_LOCATOR(
                name=name_or_object, reason=e, _meta={"type": "connector"}
            )



from spaceone.core import config
from spaceone.core.base import CoreObject

__all__ = ['BaseConnector']


class BaseConnector(CoreObject):

    name = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        connector_name = self.name or self.__class__.__name__
        self.config = config.get_connector(connector_name)



from opentelemetry import metrics
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader

from spaceone.core import config

__all__ = ['set_metric']


def set_metric():
    # Not Implemented
    # service = config.get_service()
    # endpoint = config.get_global('OTEL.endpoint')
    # _init_metric(service, endpoint)
    pass


def _init_metric(service, endpoint):
    resource = Resource(attributes={
        SERVICE_NAME: service
    })

    exporter = OTLPMetricExporter(endpoint=endpoint)
    reader = PeriodicExportingMetricReader(exporter)
    provider = MeterProvider(resource=resource, metric_readers=[reader])
    metrics.set_meter_provider(provider)



import logging
import types
import grpc
from google.protobuf.json_format import ParseDict
from google.protobuf.message_factory import MessageFactory  # , GetMessageClass
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.descriptor import ServiceDescriptor, MethodDescriptor
from grpc_reflection.v1alpha.proto_reflection_descriptor_database import (
    ProtoReflectionDescriptorDatabase,
)
from spaceone.core.error import *

_MAX_RETRIES = 2
_GRPC_CHANNEL = {}
_LOGGER = logging.getLogger(__name__)


class _ClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor,
):
    def __init__(self, options: dict, channel_key: str, request_map: dict):
        self._request_map = request_map
        self._channel_key = channel_key
        self.metadata = options.get("metadata", {})

    def _check_message(self, client_call_details, request_or_iterator, is_stream):
        if client_call_details.method in self._request_map:
            if is_stream:
                if not isinstance(request_or_iterator, types.GeneratorType):
                    raise Exception(
                        "Stream method must be specified as a generator type."
                    )

                return self._generate_message(
                    request_or_iterator, client_call_details.method
                )

            else:
                return self._make_message(
                    request_or_iterator, client_call_details.method
                )

        return request_or_iterator

    def _make_message(self, request, method_key):
        if isinstance(request, dict):
            return ParseDict(request, self._request_map[method_key]())

        else:
            return request

    def _generate_message(self, request_iterator, method_key):
        for request in request_iterator:
            yield self._make_message(request, method_key)

    def _check_error(self, response):
        if isinstance(response, Exception):
            details = response.details()
            status_code = response.code().name
            if details.startswith("ERROR_"):
                details_split = details.split(":", 1)
                if len(details_split) == 2:
                    error_code, error_message = details_split
                else:
                    error_code = details_split[0]
                    error_message = details

                if status_code == "PERMISSION_DENIED":
                    raise ERROR_PERMISSION_DENIED()
                elif status_code == "UNAUTHENTICATED":
                    raise ERROR_AUTHENTICATE_FAILURE(message=error_message)
                else:
                    e = ERROR_INTERNAL_API(message=error_message)
                    e.error_code = error_code
                    e.status_code = status_code
                    raise e

            else:
                error_message = response.details()
                if status_code == "PERMISSION_DENIED":
                    raise ERROR_PERMISSION_DENIED()
                elif status_code == "PERMISSION_DENIED":
                    raise ERROR_AUTHENTICATE_FAILURE(message=error_message)
                elif status_code == "UNAVAILABLE":
                    e = ERROR_GRPC_CONNECTION(
                        channel=self._channel_key, message=error_message
                    )
                    e.meta["channel"] = self._channel_key
                    raise e
                else:
                    e = ERROR_INTERNAL_API(message=error_message)
                    e.status_code = status_code
                    raise e

        return response

    def _generate_response(self, response_iterator):
        try:
            for response in response_iterator:
                yield self._check_error(response)
        except Exception as e:
            self._check_error(e)

    def _retry_call(
        self, continuation, client_call_details, request_or_iterator, is_stream
    ):
        retries = 0

        _LOGGER.debug(
            f"client._retry_call.init :: client_call_details: {client_call_details}"
        )

        while True:
            try:
                _LOGGER.debug(f"client._retry_call.start!!!")
                response_or_iterator = continuation(
                    client_call_details, request_or_iterator
                )
                _LOGGER.debug(f"client._retry_call.response_or_iterator Success!!!")

                if is_stream:
                    response_or_iterator = self._generate_response(response_or_iterator)
                else:
                    _LOGGER.debug(f"client._retry_call.start _check_error!!!")
                    self._check_error(response_or_iterator)
                    _LOGGER.debug(f"client._retry_call.finished _check_error!!!")

                return response_or_iterator

            except Exception as e:
                if e.error_code == "ERROR_GRPC_CONNECTION":
                    if retries >= _MAX_RETRIES:
                        channel = e.meta.get("channel")
                        if channel in _GRPC_CHANNEL:
                            _LOGGER.error(
                                f"Disconnect gRPC Endpoint. (channel = {channel})"
                            )
                            del _GRPC_CHANNEL[channel]
                        raise e
                    else:
                        _LOGGER.debug(
                            f"Retry gRPC Call: reason = {e.message}, retry = {retries + 1}"
                        )
                else:
                    raise e

            retries += 1

    def _intercept_call(
        self,
        continuation,
        client_call_details,
        request_or_iterator,
        is_request_stream,
        is_response_stream,
    ):
        new_request_or_iterator = self._check_message(
            client_call_details, request_or_iterator, is_request_stream
        )

        return self._retry_call(
            continuation,
            client_call_details,
            new_request_or_iterator,
            is_response_stream,
        )

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(
            continuation, client_call_details, request, False, False
        )

    def intercept_unary_stream(self, continuation, client_call_details, request):
        return self._intercept_call(
            continuation, client_call_details, request, False, True
        )

    def intercept_stream_unary(
        self, continuation, client_call_details, request_iterator
    ):
        return self._intercept_call(
            continuation, client_call_details, request_iterator, True, False
        )

    def intercept_stream_stream(
        self, continuation, client_call_details, request_iterator
    ):
        return self._intercept_call(
            continuation, client_call_details, request_iterator, True, True
        )


class _GRPCStub(object):
    def __init__(
        self,
        desc_pool: DescriptorPool,
        service_desc: ServiceDescriptor,
        channel: grpc.Channel,
    ):
        self._desc_pool = desc_pool
        for method_desc in service_desc.methods:
            self._bind_grpc_method(service_desc, method_desc, channel)

    def _bind_grpc_method(
        self,
        service_desc: ServiceDescriptor,
        method_desc: MethodDescriptor,
        channel: grpc.Channel,
    ):
        method_name = method_desc.name
        method_key = f"/{service_desc.full_name}/{method_name}"
        request_desc = self._desc_pool.FindMessageTypeByName(
            method_desc.input_type.full_name
        )
        request_message_desc = MessageFactory(self._desc_pool).GetPrototype(
            request_desc
        )
        # request_message_desc = GetMessageClass(request_desc)

        response_desc = self._desc_pool.FindMessageTypeByName(
            method_desc.output_type.full_name
        )
        response_message_desc = MessageFactory(self._desc_pool).GetPrototype(
            response_desc
        )
        # response_message_desc = GetMessageClass(response_desc)

        if method_desc.client_streaming and method_desc.server_streaming:
            setattr(
                self,
                method_name,
                channel.stream_stream(
                    method_key,
                    request_serializer=request_message_desc.SerializeToString,
                    response_deserializer=response_message_desc.FromString,
                ),
            )
        elif method_desc.client_streaming and not method_desc.server_streaming:
            setattr(
                self,
                method_name,
                channel.stream_unary(
                    method_key,
                    request_serializer=request_message_desc.SerializeToString,
                    response_deserializer=response_message_desc.FromString,
                ),
            )
        elif not method_desc.client_streaming and method_desc.server_streaming:
            setattr(
                self,
                method_name,
                channel.unary_stream(
                    method_key,
                    request_serializer=request_message_desc.SerializeToString,
                    response_deserializer=response_message_desc.FromString,
                ),
            )
        else:
            setattr(
                self,
                method_name,
                channel.unary_unary(
                    method_key,
                    request_serializer=request_message_desc.SerializeToString,
                    response_deserializer=response_message_desc.FromString,
                ),
            )


class GRPCClient(object):
    def __init__(self, channel, options, channel_key):
        self._request_map = {}
        self._api_resources = {}

        self._reflection_db = ProtoReflectionDescriptorDatabase(channel)
        self._desc_pool = DescriptorPool(self._reflection_db)
        self._init_grpc_reflection()

        _client_interceptor = _ClientInterceptor(
            options, channel_key, self._request_map
        )
        _intercept_channel = grpc.intercept_channel(channel, _client_interceptor)
        self._bind_grpc_stub(_intercept_channel)

    @property
    def api_resources(self):
        return self._api_resources

    def _init_grpc_reflection(self):
        for service in self._reflection_db.get_services():
            service_desc: ServiceDescriptor = self._desc_pool.FindServiceByName(service)
            service_name = service_desc.name
            for method_desc in service_desc.methods:
                method_key = f"/{service}/{method_desc.name}"
                request_desc = self._desc_pool.FindMessageTypeByName(
                    method_desc.input_type.full_name
                )
                self._request_map[method_key] = MessageFactory(
                    self._desc_pool
                ).GetPrototype(request_desc)
                # self._request_map[method_key] = GetMessageClass(request_desc)

                if service_desc.name not in self._api_resources:
                    self._api_resources[service_name] = []

                self._api_resources[service_desc.name].append(method_desc.name)

    def _bind_grpc_stub(self, intercept_channel: grpc.Channel):
        for service in self._reflection_db.get_services():
            service_desc: ServiceDescriptor = self._desc_pool.FindServiceByName(service)

            setattr(
                self,
                service_desc.name,
                _GRPCStub(self._desc_pool, service_desc, intercept_channel),
            )


def _create_secure_channel(endpoint, options):
    try:
        # cert = ssl.get_server_certificate((host, port))
        # creds = grpc.ssl_channel_credentials(str.encode(cert))
        creds = grpc.ssl_channel_credentials()
    except Exception as e:
        raise ERROR_GRPC_TLS_HANDSHAKE(reason=e)

    return grpc.secure_channel(endpoint, creds, options=options)


def _create_insecure_channel(endpoint, options):
    return grpc.insecure_channel(endpoint, options=options)


def client(endpoint=None, ssl_enabled=False, max_message_length=None, **client_opts):
    if endpoint is None:
        raise Exception("Client's endpoint is undefined.")

    _LOGGER.debug(f"pygrpc.client :: endpoint: {endpoint}")
    _LOGGER.debug(
        f"pygrpc.client :: _GRPC_CHANNEL: {_GRPC_CHANNEL}  /  is in? {endpoint in _GRPC_CHANNEL}"
    )

    if endpoint not in _GRPC_CHANNEL:
        options = []

        if max_message_length:
            options.append(("grpc.max_send_message_length", max_message_length))
            options.append(("grpc.max_receive_message_length", max_message_length))

        if ssl_enabled:
            channel = _create_secure_channel(endpoint, options)
        else:
            channel = _create_insecure_channel(endpoint, options)

        try:
            grpc.channel_ready_future(channel).result(timeout=3)
        except Exception as e:
            raise ERROR_GRPC_CONNECTION(
                channel=endpoint, message="Channel is not ready."
            )

        try:
            _GRPC_CHANNEL[endpoint] = GRPCClient(channel, client_opts, endpoint)
        except Exception as e:
            if hasattr(e, "details"):
                raise ERROR_GRPC_CONNECTION(channel=endpoint, message=e.details())
            else:
                raise ERROR_GRPC_CONNECTION(channel=endpoint, message=str(e))

    return _GRPC_CHANNEL[endpoint]


def get_grpc_method(uri_info):
    try:
        conn = client(
            endpoint=uri_info["endpoint"], ssl_enabled=uri_info["ssl_enabled"]
        )
        return getattr(getattr(conn, uri_info["service"]), uri_info["method"])

    except ERROR_BASE as e:
        raise e
    except Exception as e:
        raise ERROR_GRPC_CONFIGURATION(
            endpoint=uri_info.get("endpoint"),
            service=uri_info.get("service"),
            method=uri_info.get("method"),
        )



import time
import consul
import logging

from spaceone.core import config
from spaceone.core.error import ERROR_CONFIGURATION

__all__ = ['get_token', '_validate_token', 'Consul']

_LOGGER = logging.getLogger(__name__)

MAX_COUNT = 10
INTERVAL = 10

class Consul:
    def __init__(self, config):
        """
        Args:
          - config: connection parameter

        Example:
            config = {
                    'host': 'consul.example.com',
                    'port': 8500
                }
        """
        self.config = self._validate_config(config)

    def _validate_config(self, config):
        """
        Parameter for Consul
        - host, port=8500, token=None, scheme=http, consistency=default, dc=None, verify=True, cert=None
        """
        options = ['host', 'port', 'token', 'scheme', 'consistency', 'dc', 'verify', 'cert']
        result = {}
        for item in options:
            value = config.get(item, None)
            if value:
                result[item] = value
        return result

    def patch_token(self, key):
        """
        Args:
            key: Query key (ex. /debug/supervisor/TOKEN)

        """
        try:
            conn = consul.Consul(**self.config)
            index, data = conn.kv.get(key)
            return data['Value'].decode('ascii')

        except Exception as e:
            _LOGGER.debug(f'[patch_token] failed: {e}')
            return False

def _validate_token(token_info):
    if isinstance(token_info, dict):
        protocol = token_info['protocol']
        if protocol == 'consul':
            consul_instance = Consul(token_info['config'])
            value = False
            while value is False:
                uri = token_info['uri']
                value = consul_instance.patch_token(uri)
                _LOGGER.warn(f'[_validate_token] token: {value[:30]} uri: {uri}')
                if value:
                    break
                time.sleep(INTERVAL)

            token = value
            return token
    raise ERROR_CONFIGURATION(key=token_info)


def get_token(name='TOKEN'):
    try:
        token = config.get_global(name)
    except Exception as e:
        _LOGGER.error(f'[get_token] config error: {name}')
        raise ERROR_CONFIGURATION(key=name)

    if token == "":
        try:
            token_info = config.get_global(f'{name}_INFO')
        except Exception as e:
            _LOGGER.error(f'[get_token] config error: {name}_INFO')
            raise ERROR_CONFIGURATION(key=name)
        token = _validate_token(token_info)
    return token





#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import os
from setuptools import setup, find_packages

setup(
    name="spaceone-core",
    version=os.environ.get("PACKAGE_VERSION"),
    description="Cloudforet Core Library",
    long_description="",
    url="https://cloudforet.io/",
    author="MEGAZONE Cloud Corp.",
    author_email="admin@spaceone.dev",
    license="Apache License 2.0",
    packages=find_packages(),
    install_requires=[
        # grpc packages
        "protobuf==3.*",
        # 'grpcio',
        "grpcio-reflection",
        "google-api-core",
        "grpcio-health-checking",
        # model packages
        "pydantic==1.*",
        "mongoengine",
        # fastapi packages
        "fastapi",
        "fastapi-utils",
        "uvicorn",
        # asyncio packages
        "asyncio",
        # data parser packages
        "PyYAML",
        "jsonschema",
        # scheduler packages
        "schedule",
        "scheduler-cron",
        # cache packages
        "redis",
        "cachetools",
        # crypto(jwt) packages
        "pycryptodome",
        "jwcrypto",
        "python-jose",
        # utils packages
        "python-dateutil",
        "python-consul",
        "dnspython",
        # HTTP packages
        "requests",
        # CLI packages
        "click",
        # AWS packages
        "boto3",
        # test framework packages
        "unittest-xml-reporting",
        "factory-boy",
        "mongomock",
        # tracing packages
        "opentelemetry-api",
        "opentelemetry-sdk",
        "opentelemetry-exporter-otlp-proto-grpc",
        "opentelemetry-instrumentation-logging",
        "opentelemetry-exporter-prometheus",
    ],
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "spaceone = spaceone.core.command:cli",
        ]
    },
)



import logging


class ExcludeFilter(logging.Filter):
    def __init__(self, rules):
        self.rules = rules

    def filter(self, record):
        for _rule in self.rules:
            if getattr(record, _rule, None) in self.rules[_rule]:
                return False

        return True



import logging
from fastapi import APIRouter, Request

_LOGGER = logging.getLogger(__name__)

router = APIRouter()


def _add_mounted_apis(app, path=None):
    apis = []
    for route in app.routes:
        apis.append({'path': f'{path}{route.path}', 'name': route.name, 'method': route.methods})
    return apis


@router.get('/api/reflection')
async def api_reflection(request: Request):
    response = {'apis': []}

    for route in request.app.routes:
        if not hasattr(route, 'methods'):
            response['apis'].append({'path': route.path, 'name': route.name, 'method': []})
            response['apis'].extend(_add_mounted_apis(route.app, route.path))
        else:
            response['apis'].append({'path': route.path, 'name': route.name, 'method': route.methods})
    return response





# Database Settings
DATABASES = {
    'default': {
        # MongoDB Example
        # 'host': '<host>',
        # 'port': 27017,
        # 'db': '<db>',
        # 'username': '<user>',
        # 'password': '<password>'
    }
}

# Cache Settings
CACHES = {
    'default': {
        # Redis Example
        # 'host': '<host>',
        # 'port': 6379,
        # 'db': 0
    }
}

# Handler Configuration
HANDLERS = {
    'authentication': [
        # Default Authentication Handler
        # {
        #     'backend': 'spaceone.core.handler.authentication_handler:AuthenticationGRPCHandler',
        #     'uri': 'grpc://identity:50051/v1/Domain/get_public_key'
        # }
    ],
    'authorization': [
        # Default Authorization Handler
        # {
        #     'backend': 'spaceone.core.handler.authorization_handler:AuthorizationGRPCHandler',
        #     'uri': 'grpc://identity:50051/v1/Authorization/verify'
        # }
    ],
    'mutation': [],
    'event': []
}

# Connector Settings
CONNECTORS = {
}

# Log Settings
LOG = {
    # 'loggers': {
    #     '<package>': {
    #         'level': 'DEBUG',
    #         'handlers': ['console']
    #     }
    # },
    # 'filters': {
    #     'masking': {
    #         'rules': {
    #             'HelloWorld.say_hello': [
    #                 'name'
    #             ]
    #         }
    #     }
    # }
}



import logging

from spaceone.core.handler import BaseAuthorizationHandler
from spaceone.core.error import ERROR_PERMISSION_DENIED

_LOGGER = logging.getLogger(__name__)


class SpaceONEAuthorizationHandler(BaseAuthorizationHandler):
    def verify(
        self, params: dict, permission: str = None, role_types: list = None
    ) -> None:
        token_type = self.transaction.get_meta("authorization.token_type")
        if token_type != "SYSTEM_TOKEN":
            if user_role_type := self.transaction.get_meta("authorization.role_type"):
                if role_types:
                    self._check_role_type(user_role_type, role_types)

                    if resource_group := params.get("resource_group"):
                        self._check_resource_group(resource_group, user_role_type)

            user_permissions = (
                self.transaction.get_meta("authorization.permissions") or []
            )
            if len(user_permissions) > 0 and permission:
                self._check_permissions(user_permissions, permission)

            if user_projects := self.transaction.get_meta("authorization.projects"):
                if request_project_id := params.get("project_id"):
                    self._check_user_projects(user_projects, request_project_id)

    def _check_role_type(self, user_role_type: str, role_types: list) -> None:
        if role_types == ["USER"]:
            owner_type = self.transaction.get_meta("authorization.owner_type")
            if owner_type == "APP":
                raise ERROR_PERMISSION_DENIED()
            else:
                self.transaction.set_meta("authorization.set_user_id", True)

        elif user_role_type not in role_types:
            raise ERROR_PERMISSION_DENIED()

    @staticmethod
    def _check_resource_group(resource_group: str, user_role_type: str) -> None:
        if user_role_type == "DOMAIN_ADMIN":
            if resource_group not in ["DOMAIN", "WORKSPACE", "PROJECT"]:
                raise ERROR_PERMISSION_DENIED()
        elif user_role_type == "WORKSPACE_OWNER":
            if resource_group not in ["WORKSPACE", "PROJECT"]:
                raise ERROR_PERMISSION_DENIED()
        elif user_role_type == "WORKSPACE_MEMBER":
            if resource_group != "PROJECT":
                raise ERROR_PERMISSION_DENIED()
        elif user_role_type == "SYSTEM_ADMIN":
            if resource_group != "SYSTEM":
                raise ERROR_PERMISSION_DENIED()

    @staticmethod
    def _check_permissions(user_permissions: list, permission: str):
        for user_permission in user_permissions:
            if permission.startswith(user_permission):
                return True

        raise ERROR_PERMISSION_DENIED()

    @staticmethod
    def _check_user_projects(user_projects: list, request_project_id: str) -> None:
        if request_project_id == "*":
            return

        if request_project_id not in user_projects:
            raise ERROR_PERMISSION_DENIED()



import logging
from datetime import datetime

import six
from spaceone.api.core.v1 import query_pb2, handler_pb2
from google.protobuf import struct_pb2
from google.protobuf.empty_pb2 import Empty
from google.protobuf.timestamp_pb2 import Timestamp

__all__ = ['change_value_type', 'change_struct_type', 'change_list_value_type',
           'change_timestamp_type', 'change_empty_type']

_INT_OR_FLOAT = six.integer_types + (float,)
LOGGER = logging.getLogger(__name__)


def change_value_type(value):
    if value is None:
        return {'null_value': 'NULL_VALUE'}
    elif isinstance(value, bool):
        return {'bool_value': value}
    elif isinstance(value, six.string_types):
        return {'string_value': value}
    elif isinstance(value, _INT_OR_FLOAT):
        return {'number_value': value}
    elif isinstance(value, dict):
        change_value = {'struct_value': struct_pb2.Struct()}
        change_value['struct_value'].update(value)
        return change_value
    elif isinstance(value, list):
        change_value = {'list_value': struct_pb2.ListValue()}
        change_value['list_value'].extend(value)
        return change_value
    else:
        return value


def change_struct_type(value):
    if isinstance(value, dict):
        change_value = struct_pb2.Struct()
        change_value.update(value)
        return change_value
    else:
        return value


def change_list_value_type(value):
    if isinstance(value, list):
        change_value = struct_pb2.ListValue()
        change_value.extend(value)
        return change_value
    else:
        return value


def change_timestamp_type(value):
    if isinstance(value, datetime):
        ts = Timestamp()
        ts.FromDatetime(value)
        return ts
    elif isinstance(value, str):
        ts = Timestamp()
        ts.FromJsonString(value)
        return ts
    else:
        return value


def change_empty_type(value):
    return Empty()



import json
import logging
import time
import copy
from multiprocessing import Process
from uuid import uuid4

import schedule
from jsonschema import validate
from scheduler import Scheduler as CronSchedulerServer
from spaceone.core import queue, config
from spaceone.core.logger import set_logger
from spaceone.core.error import ERROR_CONFIGURATION
from spaceone.core.scheduler.task_schema import SPACEONE_TASK_SCHEMA

_LOGGER = logging.getLogger(__name__)


class BaseScheduler(Process):
    def __init__(self, queue, **kwargs):
        self.queue = queue
        self.config = None

        self.global_config = config.get_global()
        super().__init__()

    def push_task(self):
        # Create Task
        try:
            tasks = self.create_task()
            _LOGGER.debug(f"[push_task] task: {len(tasks)}")
        except Exception as e:
            _LOGGER.error(f"[push_task] error create_task: {e}")
            tasks = []

        for task in tasks:
            try:
                validate(task, schema=SPACEONE_TASK_SCHEMA)
                json_task = json.dumps(task)
                _LOGGER.debug(f"[push_task] Task schema: {self._remove_metadata(task)}")
                queue.put(self.queue, json_task)
            except Exception as e:
                _LOGGER.debug(f"[push_task] Task schema: {task}, {e}")

    def run(self):
        NotImplementedError("scheduler.run is not implemented")

    def create_task(self):
        NotImplementedError("scheduler.create_task is not implemented")

    def _remove_metadata(self, task):
        copied_task = copy.deepcopy(task)
        change_stages = []
        for stage in copied_task.get("stages", []):
            if "metadata" in stage:
                stage["metadata"] = "*****"
            change_stages.append(stage)

        copied_task["stages"] = change_stages
        return copied_task


class IntervalScheduler(BaseScheduler):
    def __init__(self, queue, interval):
        super().__init__(queue)
        self.config = self.parse_config(interval)

    def parse_config(self, expr):
        """expr
        format: integer (second)
        """
        try:
            if isinstance(expr, int):
                return int(expr)
        except Exception as e:
            _LOGGER.error(f"[parse_config] Wrong configraiton, {e}")

    def run(self):
        config.set_global_force(**self.global_config)

        # Enable logging configuration
        set_logger()

        schedule.every(self.config).seconds.do(self.push_task)
        while True:
            schedule.run_pending()
            time.sleep(1)


class HourlyScheduler(BaseScheduler):
    """
    HourlyScheduler starts every HH:00
    If you want to start at different minutes
    send minute like ':15' meaning every 15 minute
    """

    def __init__(self, queue, interval=1, minute=":00"):
        super().__init__(queue)
        self.config = self.parse_config(interval)
        self.minute = minute

    def parse_config(self, expr):
        """expr
        format: integer (hour)
        """
        try:
            if isinstance(expr, int):
                return int(expr)
        except Exception as e:
            _LOGGER.error(f"[parse_config] Wrong configuration, {e}")
            raise ERROR_CONFIGURATION(key="interval")

    def run(self):
        config.set_global_force(**self.global_config)

        # Enable logging configuration
        set_logger()

        # Call push_task in every hour
        schedule.every(self.config).hours.at(self.minute).do(self.push_task)
        while True:
            schedule.run_pending()
            time.sleep(1)


class CronScheduler(BaseScheduler):
    """
    cronjob: min hour day month week
    """

    def __init__(self, queue, rule):
        super().__init__(queue)
        self.config = self.parse_config(rule)

    def parse_config(self, expr):
        """exprd
        format: min hour day month week
        * * * * *
        """
        # TODO: verify format
        return expr

    def run(self):
        config.set_global_force(**self.global_config)

        # Enable logging configuration\
        set_logger()

        if self.config is False:
            # May be error format
            return
        scheduler = CronSchedulerServer(10)
        scheduler.add(f"{uuid4()}", self.config, self.push_task)
        scheduler.start()
